/*
 * Updated for 12.01
 * Java help functions. Displays function protypes for common java classes.
 * Author Jim Morris, splash@wolfman.com
 * 
 * To use just put the javahelp.e and the java*.help files in the local epsilon folder,
 * (~/.epsilon on linux) and add
 * (load-eel-from-path "javahelp.e" 2)
 * to your einit.ecm file
 * OR
 * edit javahelp.e type ALT-F3 to compile and load and then CTRL-F3 to save.
 *
 * Usage....
 * ---------
 * ALT-F1 will extract the last character typed, if it was "(" then it
 * will extract the preceding method name, then search for that method in the
 * current javahelp file, displaying the results in a popup window (or
 * in the message line if there is only one match). All
 * classes that have that public method are displayed.
 * If the last character was "." then it will extract the variable name
 * preceding the . and try to figure out what class
 * the variable is and display all the methods for that class.
 * 
 * ALT-F2 will lookup up the selected class name or prompt for a class
 * name, and show all the public methods for that class.
 * In both cases the results show the function prototype and
 * parameters.
 * There are currently three java help files, java118.help,
 * java131.help and java140.help, the default is set to java118.help and has the most
 * commonly used public methods for all java version 1.1.8 functions,
 * java131.help has the java2 version 1.3.1 functions, and java140.help
 * has java2 1.4.0 functions.
 * 
 * The variable java_version can be set to which ever version is
 * needed (118, 131 or 140). Alternatively you can create a file in the
 * working directory or a parent directory called javaversion.tag, and
 * the version will be read from that file and set the help file
 * accordingly.
 * 
 * The help files are generated by running a program called jtags with
 * the following options... jtags -i =f -L input.txt where input.txt
 * has...
 * $JDKSRC\jdk1.1.8\src\java\io
 * $JDKSRC\jdk1.1.8\src\java\lang
 * $JDKSRC\jdk1.1.8\src\java\math
 * $JDKSRC\jdk1.1.8\src\java\net
 * $JDKSRC\jdk1.1.8\src\java\text
 * $JDKSRC\jdk1.1.8\src\java\util
 * Where $JDKSRC is the java source distribution. You could add swing
 * etc if you wanted.
 * Jtags can be found at: http://www.fleiner.com/jtags/
 * and I have added a few patches to show parameters split over multiple
 * lines, just replace the file TagVector.java with my version and
 * recompile.
 *
 * @todo:
 *	
 * @bugs:
 *	Doesn't grok scope so may pick wrong type from wrong scope.
 */

#include "eel.h"

user char java_version[]= "118";
char javahelpfile[]= "java118.help";
//char javahelpfile[]= "java131.help";
//char javahelpfile[]= "java140.help";
char lastjavahelpfile[256]= "";

when_setting_java_version()
{
    if(strcmp(java_version, "118") == 0 ||
       strcmp(java_version, "131") == 0 ||
       strcmp(java_version, "140") == 0){
	sprintf(javahelpfile, "java%s.help", java_version);
	say("Setting java help file to %s", javahelpfile);
    }else{
	say("Java version must be one of: 118, 131, 140");
    }
}

int open_java_help_file()
{
    int i;
    char *path;
    if(strcmp(lastjavahelpfile, javahelpfile) != 0){ // changed since last run
	delete_buffer("-javahelp");
    }
    
    if(!exist("-javahelp")){
	char fname[FNAMELEN], dir[FNAMELEN], *d;
	int h, n;
	// see if there is a tag file telling me what version of java I am using
	if(get_buffer_directory(dir)) d= dir;
	else d= NULL;
	if(look_up_tree(fname, "javaversion.tag", d, NULL)){
	   byte buf[4];
	   h= lowopen(fname, 0);
	   n= lowread(h, buf, 3);
	   lowclose(h);
	   if(n == 3){ // override the version
	       buf[3] = 0;
	       bytes_to_chars(java_version, buf);
	       when_setting_java_version();
	   }
	}
	
	create("-javahelp");
	bufname= "-javahelp";
	path= lookpath(javahelpfile);
	if ( !path ) {
	    delete_buffer("-javahelp");
	    error ( "Can't find file %s", javahelpfile );
	}
	i = err_file_read(path, FILETYPE_AUTO);
	if (i) {			/* error reading file */
	    delete_buffer("-javahelp");
	    error ( "Can't open file %s", path);
	}
    }else{
	bufname= "-javahelp";
	point= 0;
    }
    strcpy(lastjavahelpfile, javahelpfile);
    return 1;
}

/* read a function name at point & find it in the help file, then display the function prototype */
command get_parameter_help() on cx_tab['?'], reg_tab[NUMALT(FKEY(1))]
{
    char tag[TAGLEN], *tag_pat;
    int helpbufnum;
    char *original_buffer;
    char delim[2], pat[132], s1[132], s2[132], s3[132];
    char helpstr[512];
    int p1, p2, p3, p4, p5;
    int oldpos= point;
    int cnt= 0;
    char *np;
    char wintitle[256];

    say("");
    grab(point, point-1, delim); // get the preceding character
    
    // find the function name to get help on
    tag_pat = get_mode_string_variable("tag-pattern-%s");
    if (got_bad_number || !tag_pat)
	return;
    tiled_only();
    iter = 0;
    re_search(-1, tag_pat);
    grab(point, matchstart, tag);

    if(strcmp(delim, ".") == 0){ // show methods for the class
	// find the type of the variable
	sprintf(pat, "^(<Space|Tab>|final|volatile|private|public|protected)*([a-zA-Z0-9_]+).*[ ,;](%s)<Space|Tab>*[,;=]", tag);
	if(re_search(-1, pat)){
	    p1= find_group(2, 1);
	    p2= find_group(2, 0);
	    grab(p1, p2, s1);	// the type
	}else strcpy(s1, tag); // maybe its a static
	point = oldpos;
	java_help_show_methods(s1);
	return;
    }

    // show all matching methods and parameters
    // get the javahelp file and search it
    original_buffer = bufname;
    helpbufnum= zap("-paramhelp"); // buffer to put results in

    open_java_help_file();
  
    // find each occurrence, parse out class and call and copy it to the help buffer
    sprintf(pat, "^(%s)<Tab>(.*)<Tab>/<Caret><Space|Tab>+(<^Space|Tab>.*)%%$/;", tag);
    while(re_search(1, pat)){
	p1= find_group(1, 1);
	p2= find_group(2, 1);
	p3= find_group(2, 0);
	p4= find_group(3, 1);
	p5= find_group(3, 0);
	
	grab(p1, p2-1, s1);	// the function name
	grab(p2, p3-5, s2);	// the full path to the .java file
	grab(p4, p5-1, s3);	// the function prototype

	np= rindex(s2, '\\'); // extract filename and strip the .java
	sprintf(helpstr, "%s - %s", np+1, s3);
	buf_printf(helpbufnum, "%s\n", helpstr);
	cnt++;
    }

    // now display the results
    bufname= original_buffer;
    point = oldpos;
    
    if(cnt == 0){
	say("%s Not found", tag);
	delete_buffer("-paramhelp");
	quick_abort();
    }else if(cnt == 1){
	say(helpstr);
    }else{
	sort_another("-paramhelp", 0, 0);
	sprintf(wintitle, "Matches for method: %s", tag);
	save_var _view_title = wintitle;
	view_buf(helpbufnum, 0);
	delete_buffer("-paramhelp");    
    }
}

/*
  String ts;
    final String as, fred = "ello";

String.
fred.
as.
ts.

read(

readByte(   

readLine(

arraycopy(

entrySet(
*/


/*
 * prompt for a class name & find it in the help file, then display
 * all the function prototypes for that class
 */
command get_class_functions_help() on reg_tab[NUMALT(FKEY(2))]
{
    char tag[TAGLEN];
    
    save_spot point, mark;
    
    say("");
    fix_region();

    // find the function name to get help on, either use selected
    // region or prompt for it
    if (is_highlight_on() && mark - point < sizeof(tag)){
	grab(point, mark, tag);
	highlight_off();
    }else{
	get_string(tag, "Class: ");
    }
    
    java_help_show_methods(tag);
}

// Display all function protypes for given class name
java_help_show_methods(char *tag){
    int helpbufnum;
    char *original_buffer;
    char pat[132], s1[132], s2[132], s3[132];
    char helpstr[512];
    int p1, p2, p3, p4, p5;
    int oldpos= point;
    int cnt= 0;
    char wintitle[256];
    
    // get the javahelp file and search it
    original_buffer = bufname;
    helpbufnum= zap("-paramhelp"); // buffer to put results in

    open_java_help_file();

    // find each occurrence, parse out class and call and copy it to the help buffer
    sprintf(pat, "^(.*)<Tab>(.*\\%s%%.java)<Tab>/<Caret><Space|Tab>+(<^Space|Tab>.*)%%$/;", tag);
    while(re_search(1, pat)){
	p1= find_group(1, 1);
	p2= find_group(2, 1);
	p3= find_group(2, 0);
	p4= find_group(3, 1);
	p5= find_group(3, 0);

	grab(p1, p2-1, s1);	// the function name
	grab(p2, p3-5, s2);	// the full path to the .java file
	grab(p4, p5-1, s3);	// the function prototype

	if(cnt == 0){ // show class name
	    sprintf(wintitle, "Class: %s", s2+16);
	}
	    
	sprintf(helpstr, "%s", s3);
	buf_printf(helpbufnum, "%s\n", helpstr);
	cnt++;
    }

    // now display the results
    bufname= original_buffer;
    point = oldpos;

    if(cnt == 0){
	say("%s Not found", tag);
	delete_buffer("-paramhelp");
	quick_abort();
    }else{
	save_var _view_title = wintitle;
	view_buf(helpbufnum, 0);
	delete_buffer("-paramhelp");    
    }
}
